@@ -5,11 +5,16 @@ BLE Client for BlueZ on Linux
 import asyncio
 import logging
 import os
+import sys
 import warnings
-from typing import Callable, Dict, Optional, Union, cast
+from typing import Callable, Dict, Optional, Set, Union, cast
 from uuid import UUID
 
-import async_timeout
+if sys.version_info < (3, 11):
+    from async_timeout import timeout as async_timeout
+else:
+    from asyncio import timeout as async_timeout
+
 from dbus_fast.aio import MessageBus
 from dbus_fast.constants import BusType, ErrorType, MessageType
 from dbus_fast.message import Message
@@ -25,11 +30,14 @@ from . import defs
 from .characteristic import BleakGATTCharacteristicBlueZDBus
 from .manager import get_global_bluez_manager
 from .scanner import BleakScannerBlueZDBus
-from .utils import assert_reply
+from .utils import assert_reply, get_dbus_authenticator
 from .version import BlueZFeatures
 
 logger = logging.getLogger(__name__)
 
+# prevent tasks from being garbage collected
+_background_tasks: Set[asyncio.Task] = set()
+
 
 class BleakClientBlueZDBus(BaseBleakClient):
     """A native Linux Bleak Client
@@ -38,6 +46,7 @@ class BleakClientBlueZDBus(BaseBleakClient):
 
     Args:
         address_or_ble_device (`BLEDevice` or str): The Bluetooth address of the BLE peripheral to connect to or the `BLEDevice` object representing it.
+        services: Optional list of service UUIDs that will be used.
 
     Keyword Args:
         timeout (float): Timeout for required ``BleakScanner.find_device_by_address`` call. Defaults to 10.0.
@@ -47,7 +56,12 @@ class BleakClientBlueZDBus(BaseBleakClient):
         adapter (str): Bluetooth adapter to use for discovery.
     """
 
-    def __init__(self, address_or_ble_device: Union[BLEDevice, str], **kwargs):
+    def __init__(
+        self,
+        address_or_ble_device: Union[BLEDevice, str],
+        services: Optional[Set[str]] = None,
+        **kwargs,
+    ):
         super(BleakClientBlueZDBus, self).__init__(address_or_ble_device, **kwargs)
         # kwarg "device" is for backwards compatibility
         self._adapter: Optional[str] = kwargs.get("adapter", kwargs.get("device"))
@@ -60,6 +74,8 @@ class BleakClientBlueZDBus(BaseBleakClient):
             self._device_path = None
             self._device_info = None
 
+        self._requested_services = services
+
         # D-Bus message bus
         self._bus: Optional[MessageBus] = None
         # tracks device watcher subscription
@@ -122,127 +138,171 @@ class BleakClientBlueZDBus(BaseBleakClient):
 
         manager = await get_global_bluez_manager()
 
-        # Each BLE connection session needs a new D-Bus connection to avoid a
-        # BlueZ quirk where notifications are automatically enabled on reconnect.
-        self._bus = await MessageBus(
-            bus_type=BusType.SYSTEM, negotiate_unix_fd=True
-        ).connect()
+        async with async_timeout(timeout):
+            while True:
+                # Each BLE connection session needs a new D-Bus connection to avoid a
+                # BlueZ quirk where notifications are automatically enabled on reconnect.
+                self._bus = await MessageBus(
+                    bus_type=BusType.SYSTEM,
+                    negotiate_unix_fd=True,
+                    auth=get_dbus_authenticator(),
+                ).connect()
 
-        def on_connected_changed(connected: bool) -> None:
-            if not connected:
-                logger.debug(f"Device disconnected ({self._device_path})")
+                def on_connected_changed(connected: bool) -> None:
+                    if not connected:
+                        logger.debug(f"Device disconnected ({self._device_path})")
 
-                self._is_connected = False
+                        self._is_connected = False
 
-                if self._disconnect_monitor_event:
-                    self._disconnect_monitor_event.set()
-                    self._disconnect_monitor_event = None
+                        if self._disconnect_monitor_event:
+                            self._disconnect_monitor_event.set()
+                            self._disconnect_monitor_event = None
 
-                self._cleanup_all()
-                if self._disconnected_callback is not None:
-                    self._disconnected_callback(self)
-                disconnecting_event = self._disconnecting_event
-                if disconnecting_event:
-                    disconnecting_event.set()
+                        self._cleanup_all()
+                        if self._disconnected_callback is not None:
+                            self._disconnected_callback()
+                        disconnecting_event = self._disconnecting_event
+                        if disconnecting_event:
+                            disconnecting_event.set()
 
-        def on_value_changed(char_path: str, value: bytes) -> None:
-            callback = self._notification_callbacks.get(char_path)
+                def on_value_changed(char_path: str, value: bytes) -> None:
+                    callback = self._notification_callbacks.get(char_path)
 
-            if callback:
-                callback(bytearray(value))
+                    if callback:
+                        callback(bytearray(value))
 
-        watcher = manager.add_device_watcher(
-            self._device_path, on_connected_changed, on_value_changed
-        )
-        self._remove_device_watcher = lambda: manager.remove_device_watcher(watcher)
-
-        try:
-            try:
-                #
-                # The BlueZ backend does not disconnect devices when the
-                # application closes or crashes. This can cause problems
-                # when trying to reconnect to the same device. To work
-                # around this, we check if the device is already connected.
-                #
-                # For additional details see https://github.com/bluez/bluez/issues/89
-                #
-                if not manager.is_connected(self._device_path):
-                    logger.debug("Connecting to BlueZ path %s", self._device_path)
-                    async with async_timeout.timeout(timeout):
-                        reply = await self._bus.call(
-                            Message(
-                                destination=defs.BLUEZ_SERVICE,
-                                interface=defs.DEVICE_INTERFACE,
-                                path=self._device_path,
-                                member="Connect",
-                            )
-                        )
-
-                    if (
-                        reply.message_type == MessageType.ERROR
-                        and reply.error_name == ErrorType.UNKNOWN_OBJECT.value
-                    ):
-                        raise BleakDeviceNotFoundError(
-                            self.address,
-                            f"Device with address {self.address} was not found. It may have been removed from BlueZ when scanning stopped.",
-                        )
-
-                    assert_reply(reply)
-
-                self._is_connected = True
-
-                # Create a task that runs until the device is disconnected.
-                self._disconnect_monitor_event = asyncio.Event()
-                asyncio.ensure_future(
-                    self._disconnect_monitor(
-                        self._bus, self._device_path, self._disconnect_monitor_event
-                    )
+                watcher = manager.add_device_watcher(
+                    self._device_path, on_connected_changed, on_value_changed
+                )
+                self._remove_device_watcher = lambda: manager.remove_device_watcher(
+                    watcher
                 )
 
-                #
-                # We will try to use the cache if it exists and `dangerous_use_bleak_cache`
-                # is True.
-                #
-                await self.get_services(
-                    dangerous_use_bleak_cache=dangerous_use_bleak_cache
-                )
+                self._disconnect_monitor_event = (
+                    local_disconnect_monitor_event
+                ) = asyncio.Event()
 
-                return True
-            except BaseException:
-                # Calling Disconnect cancels any pending connect request. Also,
-                # if connection was successful but get_services() raises (e.g.
-                # because task was cancelled), the we still need to disconnect
-                # before passing on the exception.
-                if self._bus:
-                    # If disconnected callback already fired, this will be a no-op
-                    # since self._bus will be None and the _cleanup_all call will
-                    # have already disconnected.
+                try:
                     try:
-                        reply = await self._bus.call(
-                            Message(
-                                destination=defs.BLUEZ_SERVICE,
-                                interface=defs.DEVICE_INTERFACE,
-                                path=self._device_path,
-                                member="Disconnect",
+                        #
+                        # The BlueZ backend does not disconnect devices when the
+                        # application closes or crashes. This can cause problems
+                        # when trying to reconnect to the same device. To work
+                        # around this, we check if the device is already connected.
+                        #
+                        # For additional details see https://github.com/bluez/bluez/issues/89
+                        #
+                        if manager.is_connected(self._device_path):
+                            logger.debug(
+                                'skipping calling "Connect" since %s is already connected',
+                                self._device_path,
+                            )
+                        else:
+                            logger.debug(
+                                "Connecting to BlueZ path %s", self._device_path
+                            )
+                            reply = await self._bus.call(
+                                Message(
+                                    destination=defs.BLUEZ_SERVICE,
+                                    interface=defs.DEVICE_INTERFACE,
+                                    path=self._device_path,
+                                    member="Connect",
+                                )
+                            )
+
+                            assert reply is not None
+
+                            if reply.message_type == MessageType.ERROR:
+                                # This error is often caused by RF interference
+                                # from other Bluetooth or Wi-Fi devices. In many
+                                # cases, retrying will connect successfully.
+                                # Note: this error was added in BlueZ 6.62.
+                                if (
+                                    reply.error_name == "org.bluez.Error.Failed"
+                                    and reply.body
+                                    and reply.body[0] == "le-connection-abort-by-local"
+                                ):
+                                    logger.debug(
+                                        "retry due to le-connection-abort-by-local"
+                                    )
+
+                                    # When this error occurs, BlueZ actually
+                                    # connected so we get "Connected" property changes
+                                    # that we need to wait for before attempting
+                                    # to connect again.
+                                    await local_disconnect_monitor_event.wait()
+
+                                    # Jump way back to the `while True:`` to retry.
+                                    continue
+
+                                if reply.error_name == ErrorType.UNKNOWN_OBJECT.value:
+                                    raise BleakDeviceNotFoundError(
+                                        self.address,
+                                        f"Device with address {self.address} was not found. It may have been removed from BlueZ when scanning stopped.",
+                                    )
+
+                            assert_reply(reply)
+
+                        self._is_connected = True
+
+                        # Create a task that runs until the device is disconnected.
+                        task = asyncio.create_task(
+                            self._disconnect_monitor(
+                                self._bus,
+                                self._device_path,
+                                local_disconnect_monitor_event,
                             )
                         )
-                        try:
-                            assert_reply(reply)
-                        except BleakDBusError as e:
-                            # if the object no longer exists, then we know we
-                            # are disconnected for sure, so don't need to log a
-                            # warning about it
-                            if e.dbus_error != ErrorType.UNKNOWN_OBJECT.value:
-                                raise
-                    except Exception as e:
-                        logger.warning(
-                            f"Failed to cancel connection ({self._device_path}): {e}"
+                        _background_tasks.add(task)
+                        task.add_done_callback(_background_tasks.discard)
+
+                        #
+                        # We will try to use the cache if it exists and `dangerous_use_bleak_cache`
+                        # is True.
+                        #
+                        await self.get_services(
+                            dangerous_use_bleak_cache=dangerous_use_bleak_cache
                         )
 
-                raise
-        except BaseException:
-            self._cleanup_all()
-            raise
+                        return True
+                    except BaseException:
+                        # Calling Disconnect cancels any pending connect request. Also,
+                        # if connection was successful but get_services() raises (e.g.
+                        # because task was cancelled), the we still need to disconnect
+                        # before passing on the exception.
+                        if self._bus:
+                            # If disconnected callback already fired, this will be a no-op
+                            # since self._bus will be None and the _cleanup_all call will
+                            # have already disconnected.
+                            try:
+                                reply = await self._bus.call(
+                                    Message(
+                                        destination=defs.BLUEZ_SERVICE,
+                                        interface=defs.DEVICE_INTERFACE,
+                                        path=self._device_path,
+                                        member="Disconnect",
+                                    )
+                                )
+                                try:
+                                    assert_reply(reply)
+                                except BleakDBusError as e:
+                                    # if the object no longer exists, then we know we
+                                    # are disconnected for sure, so don't need to log a
+                                    # warning about it
+                                    if e.dbus_error != ErrorType.UNKNOWN_OBJECT.value:
+                                        raise
+                            except Exception as e:
+                                logger.warning(
+                                    f"Failed to cancel connection ({self._device_path}): {e}"
+                                )
+
+                        raise
+                except BaseException:
+                    # this effectively cancels the disconnect monitor in case the event
+                    # was not triggered by a D-Bus callback
+                    local_disconnect_monitor_event.set()
+                    self._cleanup_all()
+                    raise
 
     @staticmethod
     async def _disconnect_monitor(
@@ -304,8 +364,7 @@ class BleakClientBlueZDBus(BaseBleakClient):
             self._bus = None
 
             # Reset all stored services.
-            self.services = BleakGATTServiceCollection()
-            self._services_resolved = False
+            self.services = None
 
     async def disconnect(self) -> bool:
         """Disconnect from the specified GATT server.
@@ -329,7 +388,7 @@ class BleakClientBlueZDBus(BaseBleakClient):
         if self._disconnecting_event:
             # another call to disconnect() is already in progress
             logger.debug(f"already in progress ({self._device_path})")
-            async with async_timeout.timeout(10):
+            async with async_timeout(10):
                 await self._disconnecting_event.wait()
         elif self.is_connected:
             self._disconnecting_event = asyncio.Event()
@@ -344,7 +403,7 @@ class BleakClientBlueZDBus(BaseBleakClient):
                     )
                 )
                 assert_reply(reply)
-                async with async_timeout.timeout(10):
+                async with async_timeout(10):
                     await self._disconnecting_event.wait()
             finally:
                 self._disconnecting_event = None
@@ -582,15 +641,14 @@ class BleakClientBlueZDBus(BaseBleakClient):
         if not self.is_connected:
             raise BleakError("Not connected")
 
-        if self._services_resolved:
+        if self.services is not None:
             return self.services
 
         manager = await get_global_bluez_manager()
 
         self.services = await manager.get_services(
-            self._device_path, dangerous_use_bleak_cache
+            self._device_path, dangerous_use_bleak_cache, self._requested_services
         )
-        self._services_resolved = True
 
         return self.services
 
