@@ -11,6 +11,7 @@ from typing import (
     List,
     NamedTuple,
     Optional,
+    Set,
     Tuple,
     Type,
 )
@@ -18,6 +19,9 @@ from typing import (
 from ..exc import BleakError
 from .device import BLEDevice
 
+# prevent tasks from being garbage collected
+_background_tasks: Set[asyncio.Task] = set()
+
 
 class AdvertisementData(NamedTuple):
     """
@@ -165,7 +169,9 @@ class BaseBleakScanner(abc.ABC):
         if inspect.iscoroutinefunction(callback):
 
             def detection_callback(s, d):
-                asyncio.ensure_future(callback(s, d))
+                task = asyncio.create_task(callback(s, d))
+                _background_tasks.add(task)
+                task.add_done_callback(_background_tasks.discard)
 
         else:
             detection_callback = callback
@@ -197,7 +203,9 @@ class BaseBleakScanner(abc.ABC):
         try:
             device, _ = self.seen_devices[address]
 
-            device.metadata = metadata
+            device.name = name
+            device._rssi = adv.rssi
+            device._metadata = metadata
         except KeyError:
             device = BLEDevice(
                 address,
@@ -212,17 +220,17 @@ class BaseBleakScanner(abc.ABC):
         return device
 
     @abc.abstractmethod
-    async def start(self):
+    async def start(self) -> None:
         """Start scanning for devices"""
         raise NotImplementedError()
 
     @abc.abstractmethod
-    async def stop(self):
+    async def stop(self) -> None:
         """Stop scanning for devices"""
         raise NotImplementedError()
 
     @abc.abstractmethod
-    def set_scanning_filter(self, **kwargs):
+    def set_scanning_filter(self, **kwargs) -> None:
         """Set scanning filter for the BleakScanner.
 
         Args:
